/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.mod.mixin.plugin.entityactivation;

import static com.google.common.base.Preconditions.checkNotNull;

import net.minecraft.entity.Entity;
import net.minecraft.entity.EnumCreatureType;
import net.minecraft.entity.IRangedAttackMob;
import net.minecraft.entity.boss.EntityDragon;
import net.minecraft.entity.boss.EntityDragonPart;
import net.minecraft.entity.boss.EntityWither;
import net.minecraft.entity.effect.EntityWeatherEffect;
import net.minecraft.entity.item.EntityEnderCrystal;
import net.minecraft.entity.item.EntityFireworkRocket;
import net.minecraft.entity.item.EntityTNTPrimed;
import net.minecraft.entity.monster.EntityMob;
import net.minecraft.entity.monster.IMob;
import net.minecraft.entity.passive.EntityAmbientCreature;
import net.minecraft.entity.passive.EntityAnimal;
import net.minecraft.entity.passive.EntityWaterMob;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.projectile.EntityFireball;
import net.minecraft.entity.projectile.EntityThrowable;
import net.minecraft.util.AxisAlignedBB;
import net.minecraft.util.BlockPos;
import net.minecraft.util.MathHelper;
import net.minecraft.world.World;
import net.minecraft.world.WorldServer;
import net.minecraft.world.chunk.Chunk;
import net.minecraftforge.common.util.FakePlayer;
import org.spongepowered.common.SpongeImpl;
import org.spongepowered.common.config.SpongeConfig;
import org.spongepowered.common.config.SpongeConfig.WorldConfig;
import org.spongepowered.common.entity.SpongeEntityType;
import org.spongepowered.common.interfaces.entity.IMixinEntity;
import org.spongepowered.common.interfaces.world.IMixinWorld;
import org.spongepowered.common.interfaces.world.IMixinWorldProvider;

import java.util.ArrayList;
import java.util.List;

public class ActivationRange {

    static AxisAlignedBB maxBB = AxisAlignedBB.func_178781_a(0, 0, 0, 0, 0, 0);
    static AxisAlignedBB miscBB = AxisAlignedBB.func_178781_a(0, 0, 0, 0, 0, 0);
    static AxisAlignedBB creatureBB = AxisAlignedBB.func_178781_a(0, 0, 0, 0, 0, 0);
    static AxisAlignedBB monsterBB = AxisAlignedBB.func_178781_a(0, 0, 0, 0, 0, 0);
    static AxisAlignedBB aquaticBB = AxisAlignedBB.func_178781_a(0, 0, 0, 0, 0, 0);
    static AxisAlignedBB ambientBB = AxisAlignedBB.func_178781_a(0, 0, 0, 0, 0, 0);
    /**
     * Initializes an entities type on construction to specify what group this
     * entity is in for activation ranges.
     *
     * @param entity Entity to get type for
     * @return group id
     */
    public static byte initializeEntityActivationType(Entity entity) {

        // account for entities that dont extend EntityMob, EntityAmbientCreature, EntityCreature
        if (((IMob.class.isAssignableFrom(entity.getClass())
                || IRangedAttackMob.class.isAssignableFrom(entity.getClass())) && (entity.getClass() != EntityMob.class))
                || entity.isCreatureType(EnumCreatureType.MONSTER, false)) {
            return 1; // Monster
        } else if (((EntityAnimal.class.isAssignableFrom(entity.getClass()) && !entity.isCreatureType(EnumCreatureType.AMBIENT, false)) || entity
                .isCreatureType(EnumCreatureType.CREATURE, false))) {
            return 2; // Creature
        } else if (EntityWaterMob.class.isAssignableFrom(entity.getClass()) || entity.isCreatureType(EnumCreatureType.WATER_CREATURE, true)) {
            return 3; // Aquatic
        } else if (EntityAmbientCreature.class.isAssignableFrom(entity.getClass()) || entity.isCreatureType(EnumCreatureType.AMBIENT, false)) {
            return 4; // Ambient
        } else {
            return 5; // Misc
        }
    }

    /**
     * These entities are excluded from Activation range checks.
     *
     * @param entity Entity to check
     * @return boolean If it should always tick.
     */
    public static boolean initializeEntityActivationState(Entity entity) {
        if (entity.field_70170_p.field_72995_K) {
            return true;
        }
        SpongeConfig.EntityActivationRangeCategory config = getActiveConfig(entity.field_70170_p).getConfig().getEntityActivationRange();

        if ((((IMixinEntity) entity).getActivationType() == 5 && config.getMiscActivationRange() == 0)
                || (((IMixinEntity) entity).getActivationType() == 4 && config.getAmbientActivationRange() == 0)
                || (((IMixinEntity) entity).getActivationType() == 3 && config.getAquaticActivationRange() == 0)
                || (((IMixinEntity) entity).getActivationType() == 2 && config.getCreatureActivationRange() == 0)
                || (((IMixinEntity) entity).getActivationType() == 1 && config.getMonsterActivationRange() == 0)
                || (entity instanceof EntityPlayer && !(entity instanceof FakePlayer))
                || entity instanceof EntityThrowable
                || entity instanceof EntityDragon
                || entity instanceof EntityDragonPart
                || entity instanceof EntityWither
                || entity instanceof EntityFireball
                || entity instanceof EntityWeatherEffect
                || entity instanceof EntityTNTPrimed
                || entity instanceof EntityEnderCrystal
                || entity instanceof EntityFireworkRocket) {
            return true;
        }

        return false;
    }

    /**
     * Utility method to grow an AABB without creating a new AABB or touching
     * the pool, so we can re-use ones we have.
     *
     * @param target The AABB to modify
     * @param source The AABB to get initial coordinates from
     * @param x The x value to expand by
     * @param y The y value to expand by
     * @param z The z value to expand by
     */
    public static void growBb(AxisAlignedBB target, AxisAlignedBB source, int x, int y, int z) {
        target.field_72340_a = source.field_72340_a - x;
        target.field_72338_b = source.field_72338_b - y;
        target.field_72339_c = source.field_72339_c - z;
        target.field_72336_d = source.field_72336_d + x;
        target.field_72337_e = source.field_72337_e + y;
        target.field_72334_f = source.field_72334_f + z;
    }

    /**
     * Find what entities are in range of the players in the world and set
     * active if in range.
     *
     * @param world The world to perform activation checks in
     */
    public static void activateEntities(World world) {
        SpongeConfig.EntityActivationRangeCategory config = getActiveConfig(world).getConfig().getEntityActivationRange();
        final int miscActivationRange = config.getMiscActivationRange();
        final int creatureActivationRange = config.getCreatureActivationRange();
        final int monsterActivationRange = config.getMonsterActivationRange();
        final int aquaticActivationRange = config.getAquaticActivationRange();
        final int ambientActivationRange = config.getAmbientActivationRange();

        int[] ranges = {miscActivationRange, creatureActivationRange, monsterActivationRange, aquaticActivationRange, ambientActivationRange};
        int maxRange = 0;
        for (int range : ranges) {
            if (range > maxRange) {
                maxRange = range;
            }
        }

        maxRange = Math.max(maxRange, miscActivationRange);
        maxRange = Math.min((6 << 4) - 8, maxRange);

        for (Object entity : world.field_73010_i) {

            Entity player = (Entity) entity;
            ((IMixinEntity) player).setActivatedTick(world.func_72912_H().func_82573_f());
            growBb(maxBB, player.func_174813_aQ(), maxRange, 256, maxRange);
            growBb(miscBB, player.func_174813_aQ(), miscActivationRange, 256, miscActivationRange);
            growBb(creatureBB, player.func_174813_aQ(), creatureActivationRange, 256, creatureActivationRange);
            growBb(monsterBB, player.func_174813_aQ(), monsterActivationRange, 256, monsterActivationRange);
            growBb(aquaticBB, player.func_174813_aQ(), aquaticActivationRange, 256, aquaticActivationRange);
            growBb(ambientBB, player.func_174813_aQ(), ambientActivationRange, 256, ambientActivationRange);

            int i = MathHelper.func_76128_c(maxBB.field_72340_a / 16.0D);
            int j = MathHelper.func_76128_c(maxBB.field_72336_d / 16.0D);
            int k = MathHelper.func_76128_c(maxBB.field_72339_c / 16.0D);
            int l = MathHelper.func_76128_c(maxBB.field_72334_f / 16.0D);

            for (int i1 = i; i1 <= j; ++i1) {
                for (int j1 = k; j1 <= l; ++j1) {
                    WorldServer worldserver = (WorldServer) world;
                    if (worldserver.field_73059_b.func_73149_a(i1, j1)) {
                        activateChunkEntities(world.func_72964_e(i1, j1));
                    }
                }
            }
        }
    }

    /**
     * Checks for the activation state of all entities in this chunk.
     *
     * @param chunk Chunk to check for activation
     */
    @SuppressWarnings("rawtypes")
    private static void activateChunkEntities(Chunk chunk) {
        for (int i = 0; i < chunk.func_177429_s().length; ++i) {

            for (Object o : chunk.func_177429_s()[i]) {
                Entity entity = (Entity) o;
                SpongeConfig<?> config = getActiveConfig(entity.field_70170_p);
                // TODO
                if (config == null) continue;
                SpongeEntityType type = (SpongeEntityType) ((org.spongepowered.api.entity.Entity) entity).getType();
                if (entity.field_70170_p.func_72912_H().func_82573_f() > ((IMixinEntity) entity).getActivatedTick()) {
                    if (((IMixinEntity) entity).getDefaultActivationState()) {
                        ((IMixinEntity) entity).setActivatedTick(entity.field_70170_p.func_72912_H().func_82573_f());
                        continue;
                    }
                    if (!config.getRootNode().getNode(SpongeConfig.MODULE_ENTITY_ACTIVATION_RANGE, type.getModId(), "enabled").getBoolean()
                            || !config.getRootNode()
                            .getNode(SpongeConfig.MODULE_ENTITY_ACTIVATION_RANGE, type.getModId(), "entities", type.getName())
                            .getBoolean()) {
                        continue;
                    }
                    switch (((IMixinEntity) entity).getActivationType()) {
                        case 1:
                            if (monsterBB.func_72326_a(entity.func_174813_aQ())) {
                                ((IMixinEntity) entity).setActivatedTick(entity.field_70170_p.func_72912_H().func_82573_f());
                            }
                            break;
                        case 2:
                            if (creatureBB.func_72326_a(entity.func_174813_aQ())) {
                                ((IMixinEntity) entity).setActivatedTick(entity.field_70170_p.func_72912_H().func_82573_f());
                            }
                            break;
                        case 3:
                            if (aquaticBB.func_72326_a(entity.func_174813_aQ())) {
                                ((IMixinEntity) entity).setActivatedTick(entity.field_70170_p.func_72912_H().func_82573_f());
                            }
                            break;
                        case 4:
                            if (ambientBB.func_72326_a(entity.func_174813_aQ())) {
                                ((IMixinEntity) entity).setActivatedTick(entity.field_70170_p.func_72912_H().func_82573_f());
                            }
                            break;
                        case 5:
                        default:
                            if (miscBB.func_72326_a(entity.func_174813_aQ())) {
                                ((IMixinEntity) entity).setActivatedTick(entity.field_70170_p.func_72912_H().func_82573_f());
                            }
                    }
                }
            }
        }
    }

    /**
     * If an entity is not in range, do some more checks to see if we should
     * give it a shot.
     *
     * @param entity Entity to check
     * @return Whether entity should still be maintained active
     */
    public static boolean checkEntityImmunities(Entity entity) {
        return false;
    }

    /**
     * Checks if the entity is active for this tick.
     *
     * @param entity The entity to check for activity
     * @return Whether the given entity should be active
     */
    public static boolean checkIfActive(Entity entity) {
        if (entity.field_70170_p.field_72995_K) {
            return true;
        }

        IMixinEntity spongeEntity = (IMixinEntity) entity;
        boolean isActive =
                spongeEntity.getActivatedTick() >= entity.field_70170_p.func_72912_H().func_82573_f() || spongeEntity.getDefaultActivationState();

        // Should this entity tick?
        if (!isActive) {
            if ((entity.field_70170_p.func_72912_H().func_82573_f() - spongeEntity.getActivatedTick() - 1) % 20 == 0) {
                // Check immunities every 20 ticks.
                if (checkEntityImmunities(entity)) {
                    // Triggered some sort of immunity, give 20 full ticks before we check again.
                    spongeEntity.setActivatedTick(entity.field_70170_p.func_72912_H().func_82573_f() + 20);
                }
                isActive = true;
            }
            // Add a little performance juice to active entities. Skip 1/4 if not immune.
        } else if (!spongeEntity.getDefaultActivationState() && entity.field_70173_aa % 4 == 0 && !checkEntityImmunities(entity)) {
            isActive = false;
        }

        // Make sure not on edge of unloaded chunk
        int x = MathHelper.func_76128_c(entity.field_70165_t);
        int z = MathHelper.func_76128_c(entity.field_70161_v);
        if (isActive && !entity.field_70170_p.func_175697_a(new BlockPos(x, 0, z), 16)) {
            isActive = false;
        }

        return isActive;
    }

    public static void addEntityToConfig(World world, SpongeEntityType type, byte activationType) {
        checkNotNull(world, "world");
        checkNotNull(type, "type");

        List<SpongeConfig<?>> configs = new ArrayList<>();
        configs.add(SpongeImpl.getGlobalConfig());
        configs.add(((IMixinWorldProvider) world.field_73011_w).getDimensionConfig());
        configs.add(((IMixinWorld) world).getWorldConfig());
        checkNotNull(configs.get(0), "global");
        checkNotNull(configs.get(1), "dimension");
        checkNotNull(configs.get(2), "world");
        String entityType = "misc";
        if (activationType == 1) {
            entityType = "monster";
        } else if (activationType == 2) {
            entityType = "creature";
        } else if (activationType == 3) {
            entityType = "aquatic";
        } else if (activationType == 4) {
            entityType = "ambient";
        }

        for (SpongeConfig<?> config : configs) {
            // TODO
            if (config == null) continue;
            if (config.getRootNode().getNode(SpongeConfig.MODULE_ENTITY_ACTIVATION_RANGE, type.getModId()).isVirtual()) {
                config.getRootNode().getNode(SpongeConfig.MODULE_ENTITY_ACTIVATION_RANGE, type.getModId(), "enabled").setValue(true);
            }

            if (config.getRootNode().getNode(SpongeConfig.MODULE_ENTITY_ACTIVATION_RANGE, type.getModId(), entityType, type.getName())
                    .isVirtual()) {
                config.getRootNode().getNode(SpongeConfig.MODULE_ENTITY_ACTIVATION_RANGE, type.getModId(), entityType, type.getName())
                        .setValue(true);
                config.save();
            }
        }
    }

    public static SpongeConfig<?> getActiveConfig(World world) {
        SpongeConfig<WorldConfig> config = ((IMixinWorld) world).getWorldConfig();
        if (config.getConfig().isConfigEnabled()) {
            return config;
        } else if (((IMixinWorldProvider) world.field_73011_w).getDimensionConfig() != null && ((IMixinWorldProvider) world.field_73011_w)
                .getDimensionConfig().getConfig().isConfigEnabled()) {
            return ((IMixinWorldProvider) world.field_73011_w).getDimensionConfig();
        } else {
            return SpongeImpl.getGlobalConfig();
        }
    }
}
